/**
 * Copyright (C) 2024 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License.  You may obtain a copy
 * of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include "collect_global_pragma.h"

#include <algorithm>

#include "backends/tofino/bf-p4c/common/pragma/all_pragmas.h"
#include "backends/tofino/bf-p4c/parde/clot/pragma/do_not_use_clot.h"
#include "backends/tofino/bf-p4c/phv/pragma/phv_pragmas.h"

const std::vector<std::string> *CollectGlobalPragma::g_global_pragma_names =
    new std::vector<std::string>{PragmaAlias::name,
                                 PragmaAtomic::name,
                                 PragmaAutoInitMetadata::name,
                                 PragmaBackwardCompatible::name,
                                 PragmaBytePack::name,
                                 PragmaContainerSize::name,
                                 PragmaContainerType::name,
                                 PragmaDisableI2EReservedDropImplementation::name,
                                 PragmaDisableEgressMirrorIOSelectInitialization::name,
                                 PragmaDoNotUseClot::name,
                                 PragmaEgressIntrinsicMetadataOpt::name,
                                 PragmaGFMParityEnable::name,
                                 PragmaMutuallyExclusive::name,
                                 PragmaNoInit::name,
                                 PragmaNoOverlay::name,
                                 PragmaNoPack::name,
                                 PragmaParserBandwidthOpt::name,
                                 PragmaParserGroupMonogress::name,
                                 PragmaAllowPOVonMocha::name,
                                 PragmaPhvLimit::name,
                                 PragmaPrioritizeARAinits::name,
                                 PragmaQuickPhvAlloc::name,
                                 PragmaSolitary::name,
                                 PHV::pragma::DISABLE_DEPARSE_ZERO};

cstring CollectGlobalPragma::getStructFieldName(const IR::StructField *s) const {
    const auto nameAnnotation = s->getAnnotation(IR::Annotation::nameAnnotation);
    if (!nameAnnotation || nameAnnotation->getExpr().size() != 1) return cstring();
    auto structName = nameAnnotation->getExpr(0)->to<IR::StringLiteral>();
    if (!structName) return cstring();
    return structName->value;
}

bool CollectGlobalPragma::preorder(const IR::StructField *s) {
    cstring structFieldName = getStructFieldName(s);
    // The header names are prefixed with a '.'. For the purposes of PHV allocation, we do not need
    // the period.
    cstring structFieldNameWithoutDot =
        (structFieldName != cstring()) ? structFieldName.substr(1) : structFieldName;
    for (auto ann : s->annotations) {
        // Ignore annotations that are not NOT_PARSED or NOT_DEPARSED
        bool notParsedDeparsedAnnotation = (ann->name.name == PHV::pragma::NOT_PARSED ||
                                            ann->name.name == PHV::pragma::NOT_DEPARSED);
        if (!notParsedDeparsedAnnotation) continue;
        if (structFieldName == cstring())
            BUG("Pragma %1% on Struct Field %2% without a name", ann->name.name, s);
        // For the notParsedDeparsedAnnotation, create a new annotation that includes
        // structFieldName.
        if (!ann->getExpr().size()) continue;

        auto &exprs = ann->getExpr();

        const unsigned min_required_arguments = 1;  // gress
        unsigned required_arguments = min_required_arguments;
        unsigned expr_index = 0;
        const IR::StringLiteral *pipe_arg = nullptr;
        const IR::StringLiteral *gress_arg = nullptr;

        if (!PHV::Pragmas::determinePipeGressArgs(exprs, expr_index, required_arguments, pipe_arg,
                                                  gress_arg)) {
            continue;
        }

        if (!PHV::Pragmas::checkNumberArgs(ann, required_arguments, min_required_arguments, true,
                                           ann->name.name, "`gress'"_cs)) {
            continue;
        }

        // Create a new annotation that includes the pragma not_parsed or not_deparsed and also adds
        // the name of the header with which the pragma is associated. The format is equivalent to:
        // @pragma not_parsed <gress> <hdr_name>. PHV allocation can then consume this added
        // annotation when implementing the deparsed zero optimization.
        auto *name = new IR::StringLiteral(IR::ID(structFieldNameWithoutDot));
        IR::Annotation *newAnnotation =
            new IR::Annotation(IR::ID(ann->name.name), {pipe_arg, gress_arg, name});
        LOG1("      Adding global annotation: " << newAnnotation);
        global_pragmas_.push_back(newAnnotation);
    }
    return true;
}

bool CollectGlobalPragma::preorder(const IR::Annotation *annotation) {
    auto pragma_name = annotation->name.name;
    bool is_global_pragma = (std::find(g_global_pragma_names->begin(), g_global_pragma_names->end(),
                                       pragma_name) != g_global_pragma_names->end());
    if (is_global_pragma) {
        global_pragmas_.push_back(annotation);
        LOG1("      Adding global annotation: " << annotation);
    }
    return false;
}

const IR::Annotation *CollectGlobalPragma::exists(const char *pragma_name) const {
    const IR::Annotation *pragma = nullptr;
    for (auto p : global_pragmas()) {
        if (p->name == pragma_name) {
            pragma = p;
            break;
        }
    }
    return pragma;
}
