/**
 * Copyright (C) 2024 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License.  You may obtain a copy
 * of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 *
 * SPDX-License-Identifier: Apache-2.0
 */

/**
 *  Collect pipelines form program.
 */
#include "collect_pipelines.h"

#include "backends/tofino/bf-p4c/device.h"

namespace BFN {

template <typename T>
static bool ptrNameEq(T *a, T *b) {
    return a == b || (a && b && a->getName() == b->getName());
}

bool CollectPipelines::FullGress::operator==(const FullGress &other) const {
    return ptrNameEq(parser, other.parser) && ptrNameEq(control, other.control) &&
           ptrNameEq(deparser, other.deparser);
}

template <typename Type>
static void _setPipe(CollectPipelines::Pipe *self, const IR::IDeclaration *dec,
                     CollectPipelines::FullGress CollectPipelines::Pipe::*gress,
                     Type *CollectPipelines::FullGress::*gressField, cstring type) {
    auto val = dec->to<Type>();
    BUG_CHECK(val, "invalid type of %1%, expecting %2%", dec, type);
    (self->*gress).*gressField = val;
}

void CollectPipelines::Pipe::set(unsigned count, unsigned idx, const IR::IDeclaration *dec) {
    using Gress = CollectPipelines::FullGress;

    BUG_CHECK(idx < count, "Pipe argument %1% out of range", idx);

    BUG_CHECK(count == 6u || count == 7u, "Cannot process pipelines with %1% arguments", count);
    if (idx == 6) {
        ghost = dec->to<IR::BFN::TnaControl>();
        BUG_CHECK(ghost, "invalid type of %1%, expecting control", dec);
    } else {
        auto gress = idx < 3 ? &Pipe::ingress : &Pipe::egress;
        idx %= 3;
        switch (idx) {
            case 0:
                _setPipe(this, dec, gress, &Gress::parser, "parser"_cs);
                break;
            case 1:
                _setPipe(this, dec, gress, &Gress::control, "control"_cs);
                break;
            case 2:
                _setPipe(this, dec, gress, &Gress::deparser, "deparser"_cs);
                break;
        }
    }
}

bool CollectPipelines::Pipe::operator==(const Pipe &other) const {
    return ptrNameEq(dec, other.dec);
}

bool CollectPipelines::Pipe::operator<(const Pipe &other) const {
    return (dec && other.dec && dec->getName().name == other.dec->getName().name) ||
           dec < other.dec;
}

// Checks the "main"
bool CollectPipelines::preorder(const IR::Declaration_Instance *di) {
    // Check if this is specialized type with baseType referring to "Switch" or
    // "Pipeline"
    auto type = di->type->to<IR::Type_Specialized>();
    if (!type || !type->baseType->is<IR::Type_Name>()) return false;

    auto name = type->baseType->to<IR::Type_Name>()->path->name;
    if (name == "Switch") {
        for (auto argument : *di->arguments) {
            auto pathExp = argument->expression->to<IR::PathExpression>();
            BUG_CHECK(pathExp && pathExp->type->is<IR::Type_Package>(),
                      "pipe expression %1% has unexpected type", argument->expression);

            auto pipe = pipesByDec.at(refMap->getDeclaration(pathExp->path));
            pipelines->pipes.push_back(pipe);
            const IR::Type_Declaration *decs[]{
                pipe.ingress.parser, pipe.ingress.control, pipe.ingress.deparser,
                pipe.egress.parser,  pipe.egress.control,  pipe.egress.deparser,
                pipe.ghost};
            for (auto dec : decs) {
                if (dec) pipelines->declarationToPipe[dec].insert(pipelines->pipes.size() - 1);
            }
        }
        LOG6("[CollectPipelines] Found switch with " << di->arguments->size() << " pipes");
    }

    if (name == "Pipeline") {
        unsigned i = 0;
        Pipe pipe;
        for (auto *argument : *di->arguments) {
            auto *call = argument->expression->to<IR::ConstructorCallExpression>();
            BUG_CHECK(call, "Cannot handle pipelines with non-call expression %1% (in %2%)",
                      argument->expression, di);
            BUG_CHECK(call->constructedType, "Cannot get the callee in expression %1% (in %2%)",
                      call, di);
            auto *type = call->constructedType->to<IR::Type_Name>();
            BUG_CHECK(type, "Cannot extract type from type name in expression %1% (in %2%)", call,
                      di);
            auto *dec = refMap->getDeclaration(type->path);
            pipe.set(di->arguments->size(), i, dec);
            ++i;
        }
        pipe.dec = di;
        pipesByDec[di] = pipe;
        LOG6("[CollectPipelines] Found pipeline " << di->getName() << " "
                                                  << (pipe.ghost ? "with ghost" : "without ghost"));
    }
    return false;
}

}  // namespace BFN
