/*
Copyright 2013-present Barefoot Networks, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "constantParsing.h"

#include "frontends/common/options.h"
#include "ir/ir.h"
#include "lib/big_int_util.h"
#include "lib/source_file.h"

namespace P4 {

std::ostream &operator<<(std::ostream &out, const UnparsedConstant &constant) {
    out << "UnparsedConstant(" << constant.text << ',' << constant.skip << ',' << constant.base
        << ',' << constant.hasWidth << ')';
    return out;
}

bool operator<(const UnparsedConstant &a, const UnparsedConstant &b) {
    return std::tie(a.text, a.skip, a.base, a.hasWidth) <
           std::tie(b.text, b.skip, b.base, b.hasWidth);
}

/// A helper to parse constants which have an explicit width;
/// @see UnparsedConstant for an explanation of the parameters.
static IR::Constant *parseConstantWithWidth(Util::SourceInfo srcInfo, const char *text,
                                            unsigned skip, unsigned base) {
    char *sep;
    auto size = strtol(text, &sep, 10);
    sep += strspn(sep, " \t\r\n");
    if (!*sep) BUG("Expected to find separator %1%", text);
    if (size < 0) {
        ::P4::error(ErrorType::ERR_INVALID, "%1%: invalid width; %2% must be positive", srcInfo,
                    size);
        return nullptr;
    }
    if (size > P4CContext::getConfig().maximumWidthSupported()) {
        ::P4::error(ErrorType::ERR_OVERLIMIT, "%1%: %2% size too large", srcInfo, size);
        return nullptr;
    }

    bool isSigned = *sep++ == 's';
    sep += strspn(sep, " \t\r\n");
    big_int value = Util::cvtInt(sep + skip, base);
    const IR::Type *type = IR::Type_Bits::get(srcInfo, size, isSigned);
    IR::Constant *result = new IR::Constant(srcInfo, type, value, base);
    return result;
}

IR::Constant *parseConstant(const Util::SourceInfo &srcInfo, const UnparsedConstant &constant,
                            long defaultValue) {
    if (!constant.hasWidth) {
        auto value = Util::cvtInt(constant.text.c_str() + constant.skip, constant.base);
        return new IR::Constant(srcInfo, value, constant.base);
    }

    auto result =
        parseConstantWithWidth(srcInfo, constant.text.c_str(), constant.skip, constant.base);
    if (result == nullptr) return new IR::Constant(srcInfo, defaultValue);
    return result;
}

int parseConstantChecked(const Util::SourceInfo &srcInfo, const UnparsedConstant &constant) {
    auto cst = parseConstant(srcInfo, constant, 0);
    if (!cst->fitsInt()) {
        ::P4::error(ErrorType::ERR_OVERLIMIT,
                    "%1$x: this implementation does not support bitstrings this large", cst);
        return 8;  // this is a fine value for a width; compilation will stop anyway
    }
    return cst->asInt();
}

}  // namespace P4
